@@ -16,7 +16,7 @@ from .circuit import Circuit, BlueqatGlobalSetting
 from . import pauli
 from . import utils
 from . import vqe
-from . import opt
+from . import wq
 from ._version import __version__
 
-__all__ = ["pauli", "utils", "vqe", "opt", "Circuit", "BlueqatGlobalSetting"]
+__all__ = ["pauli", "utils", "vqe", "wq", "Circuit", "BlueqatGlobalSetting"]
