@@ -100,7 +100,7 @@ def _mult_shifted(masks, idx):
 def _zgate(qubits, n_qubits, target):
     lower_mask = (1 << _QSMask(target)) - 1
     for i in prange(1 << (_QSMask(n_qubits) - 1)):
-        qubits[_shifted(lower_mask, i)] *= -1
+        qubits[_shifted(lower_mask, i) + (1 << target)] *= -1
 
 
 @njit(locals={'lower_mask': _QSMask},
@@ -120,9 +120,8 @@ def _ygate(qubits, n_qubits, target):
     lower_mask = (1 << _QSMask(target)) - 1
     for i in prange(1 << (_QSMask(n_qubits) - 1)):
         i0 = _shifted(lower_mask, i)
-        t = qubits[i0]
-        # Global phase is ignored.
-        qubits[i0] = -qubits[i0 + (1 << target)]
+        t = qubits[i0] * 1j
+        qubits[i0] = qubits[i0 + (1 << target)] * -1j
         qubits[i0 + (1 << target)] = t
 
 
@@ -145,7 +144,6 @@ def _diaggate(qubits, n_qubits, target, factor):
     lower_mask = (1 << _QSMask(target)) - 1
     for i in prange(1 << (_QSMask(n_qubits) - 1)):
         i1 = _shifted(lower_mask, i) + (1 << target)
-        # Global phase is ignored.
         qubits[i1] *= factor
 
 
@@ -347,6 +345,16 @@ def _reduce1(qubits, target, n_qubits, p0):
         qubits[i0] = 0.0
 
 
+@njit(nogil=True, parallel=True, fastmath=FASTMATH)
+def _reset1(qubits, target, n_qubits, p0):
+    sqrtp_inv = 1.0 / math.sqrt(1.0 - p0)
+    lower_mask = (1 << _QSMask(target)) - 1
+    for i in prange(1 << (_QSMask(n_qubits) - 1)):
+        i0 = _shifted(lower_mask, i)
+        qubits[i0] = qubits[i0 + (1 << target)] * sqrtp_inv
+        qubits[i0 + (1 << target)] = 0.0
+
+
 class NumbaBackend(Backend):
     """Simulator backend which uses numba."""
     __return_type = {
@@ -377,7 +385,7 @@ class NumbaBackend(Backend):
             return
 
     def run(self, gates, n_qubits, *args, **kwargs):
-        def __parse_run_args(shots=None, returns=None, enable_cache=True, ignore_global=True,
+        def __parse_run_args(shots=None, returns=None, enable_cache=True, ignore_global=False,
                              dtype=DEFAULT_DTYPE, **_kwargs):
             if returns is None:
                 if shots is None:
@@ -618,6 +626,21 @@ class NumbaBackend(Backend):
                 ctx.cregs[target] = 1
         return ctx
 
+    def gate_reset(self, gate, ctx):
+        if ctx.save_cache:
+            self.cache = ctx.qubits.copy()
+        ctx.save_cache = False
+        qubits = ctx.qubits
+        n_qubits = ctx.n_qubits
+        for target in gate.target_iter(n_qubits):
+            rand = random.random()
+            p0 = _p0calc(qubits, target, n_qubits)
+            if rand < p0:
+                _reduce0(qubits, target, n_qubits, p0)
+            else:
+                _reset1(qubits, target, n_qubits, p0)
+        return ctx
+
 
 @njit(nogil=True, cache=True)
 def _ignore_global(qubits):
