@@ -14,9 +14,10 @@
 
 """The module for calculate Pauli matrices."""
 
+from bisect import bisect_left
 from collections import defaultdict, namedtuple
 from functools import reduce
-from itertools import combinations, product
+from itertools import combinations, groupby, product
 from numbers import Number, Integral
 from math import pi
 
@@ -26,6 +27,7 @@ import scipy.sparse
 _PauliTuple = namedtuple("_PauliTuple", "n")
 half_pi = pi / 2
 
+
 _sparse_types = {
         'bsr': scipy.sparse.bsr_matrix,
         'coo': scipy.sparse.coo_matrix,
@@ -36,18 +38,92 @@ _sparse_types = {
         'lil': scipy.sparse.lil_matrix,
 }
 
+
 _matrix = {
-    'I': np.array([[1, 0], [0, 1]], dtype=complex),
-    'X': np.array([[0, 1], [1, 0]], dtype=complex),
-    'Y': np.array([[0, -1j], [1j, 0]], dtype=complex),
-    'Z': np.array([[1, 0], [0, -1]], dtype=complex)
+        'I': np.array([[1, 0], [0, 1]], dtype=complex),
+        'X': np.array([[0, 1], [1, 0]], dtype=complex),
+        'Y': np.array([[0, -1j], [1j, 0]], dtype=complex),
+        'Z': np.array([[1, 0], [0, -1]], dtype=complex)
 }
 
+
+_mul_map = {
+    ('X', 'X'): (1.0, 'I'),
+    ('X', 'Y'): (1j, 'Z'),
+    ('X', 'Z'): (-1j, 'Y'),
+    ('Y', 'X'): (-1j, 'Z'),
+    ('Y', 'Y'): (1.0, 'I'),
+    ('Y', 'Z'): (1j, 'X'),
+    ('Z', 'X'): (1j, 'Y'),
+    ('Z', 'Y'): (-1j, 'X'),
+    ('Z', 'Z'): (1.0, 'I'),
+}
+
+
 _sparse_matrix = {
     ty: {ch: fn(mat, dtype=complex) for ch, mat in _matrix.items()}
     for ty, fn in _sparse_types.items()
 }
 
+
+def _kron_1d(a, b):
+    """This function is for internal use.
+    Returns a⊗b for 1d array.
+    """
+    nb = b.size
+    d = np.repeat(a, nb).reshape(-1, nb)
+    d *= b
+    return d.reshape(-1)
+
+
+def _kron_1d_rec(krons, cumsum, lo, hi):
+    """This function is for internal use.
+    Equivalent with reduce(_kron_1d, krons[lo:hi]), but faster.
+    """
+    if hi - lo == 1:
+        return krons[lo]
+    if hi - lo == 2:
+        return _kron_1d(krons[lo], krons[lo + 1])
+    mid = bisect_left(cumsum, (cumsum[lo] + cumsum[hi - 1]) // 2, lo, hi)
+    if mid == lo:
+        return _kron_1d(krons[lo], _kron_1d_rec(krons, cumsum, lo + 1, hi))
+    return _kron_1d(_kron_1d_rec(krons, cumsum, lo, mid), _kron_1d_rec(krons, cumsum, mid, hi))
+
+
+def _term_to_dataarray(term, n_qubits, rowmajor):
+    """This function is for internal use.
+    Make data of sparse Kronecker product matrix."""
+    y_mat = np.array([-1j, 1j]) if rowmajor else np.array([1j, -1j])
+    paulis = ['I'] * n_qubits
+    data_list = []
+    for op in term.ops:
+        paulis[op.n] = op.op
+    for g, l in groupby(paulis):
+        n = len(tuple(l))
+        if g == 'Y':
+            data_list += [y_mat.copy() for _ in range(n)]
+        elif g == 'Z':
+            data_list += [np.array([1, -1], dtype=complex) for _ in range(n)]
+        else:
+            data_list.append(np.repeat(np.array([1], dtype=complex), 2 ** n))
+    t = min(data_list, key=len)
+    t *= term.coeff
+    data_list.reverse()
+    cumsum = np.array([k.size for k in data_list]).cumsum()
+    return _kron_1d_rec(data_list, cumsum, 0, len(cumsum))
+
+
+def _term_to_indices(term, dim, dtype, rowcol):
+    """This function is for internal use.
+    Make indices for sparse Kronecker product matrix."""
+    xor_bits = sum(1 << op.n for op in term.ops if op.op in 'XY')
+    if rowcol:
+        col = np.arange(dim, dtype=dtype)
+        row = col ^ xor_bits
+        return row, col
+    return np.arange(dim, dtype=dtype) ^ xor_bits
+
+
 def pauli_from_char(ch, n=0):
     """Make Pauli matrix from an character.
 
@@ -72,6 +148,7 @@ def pauli_from_char(ch, n=0):
         return Z(n)
     raise ValueError("ch shall be X, Y, Z or I")
 
+
 def term_from_chars(chars):
     """Make Pauli's Term from chars which is written by "X", "Y", "Z" or "I".
     e.g. "XZIY" => X(3) * Z(2) * Y(0)
@@ -87,6 +164,7 @@ def term_from_chars(chars):
     """
     return Term.from_chars(reversed(chars))
 
+
 def to_term(pauli):
     """Convert to Term from Pauli operator (X, Y, Z, I).
 
@@ -98,6 +176,7 @@ def to_term(pauli):
     """
     return pauli.to_term()
 
+
 def to_expr(term):
     """Convert to Expr from Term or Pauli operator (X, Y, Z, I).
 
@@ -109,6 +188,7 @@ def to_expr(term):
     """
     return term.to_expr()
 
+
 def commutator(expr1, expr2):
     """Returns [expr1, expr2] = expr1 * expr2 - expr2 * expr1.
 
@@ -123,6 +203,7 @@ def commutator(expr1, expr2):
     expr2 = expr2.to_expr().simplify()
     return (expr1 * expr2 - expr2 * expr1).simplify()
 
+
 def is_commutable(expr1, expr2, eps=0.00000001):
     """Test whether expr1 and expr2 are commutable.
 
@@ -137,6 +218,7 @@ def is_commutable(expr1, expr2, eps=0.00000001):
     """
     return sum((x * x.conjugate()).real for x in commutator(expr1, expr2).coeffs()) < eps
 
+
 # To avoid pylint error
 def _n(pauli):
     return pauli.n
@@ -155,6 +237,11 @@ class _PauliImpl:
         """If `self` is I, returns True, otherwise False."""
         return self.op == "I"
 
+    @property
+    def n_qubits(self):
+        """Returns `self.n + 1` if self is not I. otherwise 0."""
+        return 0 if self.is_identity else _n(self) + 1
+
     def __hash__(self):
         return hash((self.op, _n(self)))
 
@@ -225,7 +312,6 @@ class _PauliImpl:
         """Convert to Pauli Expr"""
         return self.to_term().to_expr()
 
-
     @property
     def matrix(self):
         """Matrix reprentation of this operator."""
@@ -233,29 +319,7 @@ class _PauliImpl:
 
     def to_matrix(self, n_qubits=-1, *, sparse=None):
         """Convert to the matrix."""
-        if sparse is None:
-            kron = np.kron
-            mat = _matrix[self.op]
-            eye = np.eye
-        else:
-            if sparse not in _sparse_types:
-                raise ValueError(f'Unknown sparse format {sparse}.')
-            kron = lambda a, b: scipy.sparse.kron(a, b, format=sparse)
-            mat = _sparse_matrix[sparse][self.op]
-            eye = lambda n: scipy.sparse.eye(n, format=sparse)
-        if self.is_identity:
-            if n_qubits == -1:
-                return eye(2)
-            return eye(2 ** n_qubits)
-        if n_qubits == -1:
-            n_qubits = _n(self) + 1
-        krons = []
-        if _n(self) > 0:
-            krons.append(eye(2 ** (_n(self))))
-        krons.append(mat)
-        if n_qubits > _n(self) + 1:
-            krons.append(eye(2 ** (n_qubits - _n(self) - 1)))
-        return reduce(kron, reversed(krons))
+        return self.to_term().to_matrix(n_qubits, sparse=sparse)
 
 class _X(_PauliImpl, _PauliTuple):
     """Pauli's X operator"""
@@ -276,6 +340,11 @@ class _PauliCtor:
     def __getitem__(self, n):
         return self.ty(n)
 
+    @property
+    def matrix(self):
+        """Matrix reprentation of this operator."""
+        return _matrix[self.ty.__name__[-1]].copy()
+
 X = _PauliCtor(_X)
 Y = _PauliCtor(_Y)
 Z = _PauliCtor(_Z)
@@ -285,6 +354,11 @@ class _I(_PauliImpl, namedtuple("_I", "")):
     def __call__(self):
         return self
 
+    @property
+    def matrix(self):
+        """Matrix reprentation of this operator."""
+        return _matrix['I'].copy()
+
 I = _I()
 _TermTuple = namedtuple("_TermTuple", "ops coeff")
 
@@ -452,14 +526,7 @@ class Term(_TermTuple):
                 return 1.0, op2
             if op2 == "I":
                 return 1.0, op1
-            if op1 == op2:
-                return 1.0, "I"
-            if op1 == "X":
-                return (-1j, "Z") if op2 == "Y" else (1j, "Y")
-            if op1 == "Y":
-                return (-1j, "X") if op2 == "Z" else (1j, "Z")
-            if op1 == "Z":
-                return (-1j, "Y") if op2 == "X" else (1j, "X")
+            return _mul_map[op1, op2]
 
         before = defaultdict(list)
         for op in self.ops:
@@ -489,8 +556,19 @@ class Term(_TermTuple):
         return (op.n for op in self.ops)
 
     def max_n(self):
-        """Returns the maximum index of Pauli matrices in the Term."""
-        return max(self.n_iter())
+        """Returns the maximum index of Pauli matrices in the Term.
+        If there's no Pauli matrices, returns -1.
+        """
+        try:
+            return max(self.n_iter())
+        except ValueError:
+            return -1
+
+    @property
+    def n_qubits(self):
+        """Returns the number of qubits of the term.
+        If the term is constant with identity matrix, n_qubits is 0."""
+        return self.max_n() + 1
 
     def append_to_circuit(self, circuit, simplify=True):
         """Append Pauli gates to `Circuit`."""
@@ -539,29 +617,30 @@ class Term(_TermTuple):
 
     def to_matrix(self, n_qubits=-1, *, sparse=None):
         """Convert to the matrix."""
-        if sparse is None:
-            kron = np.kron
-            mat = _matrix
-            eye = np.eye
-        else:
-            if sparse not in _sparse_types:
-                raise ValueError(f'Unknown sparse format {sparse}.')
-            kron = lambda a, b: scipy.sparse.kron(a, b, format=sparse)
-            mat = _sparse_matrix[sparse]
-            eye = lambda n: scipy.sparse.eye(n, format=sparse)
+        if not (sparse is None or sparse in _sparse_types):
+            raise ValueError(f'Unknown sparse format {sparse}.')
         if n_qubits == -1:
-            n_qubits = self.max_n() + 1
-        simp = self.simplify()
-        n_last = -1
-        krons = []
-        for op in simp.ops:
-            if op.n > n_last + 1:
-                krons.append(eye(2 ** (op.n - n_last - 1)))
-            krons.append(mat[op.op])
-            n_last = op.n
-        if n_qubits - 1 > n_last:
-            krons.append(eye(2 ** (n_qubits - 1 - n_last)))
-        return self.coeff * reduce(kron, reversed(krons))
+            n_qubits = self.n_qubits
+        if n_qubits == 0:
+            m = np.array([[self.coeff]])
+            if sparse is None:
+                return m
+            return _sparse_types[sparse](m)
+        dim = 2 ** n_qubits
+        term = self.simplify()
+        data = _term_to_dataarray(term, n_qubits, sparse == 'csr')
+        dtype_idx = np.int32 if n_qubits < 31 else np.int64
+        if sparse == 'csc':
+            indices = _term_to_indices(term, dim, dtype_idx, False)
+            return scipy.sparse.csc_matrix((data, indices, np.arange(dim + 1, dtype=dtype_idx)), shape=(dim, dim))
+        if sparse == 'csr':
+            indices = _term_to_indices(term, dim, dtype_idx, False)
+            return scipy.sparse.csr_matrix((data, indices, np.arange(dim + 1, dtype=dtype_idx)), shape=(dim, dim))
+        row, col = _term_to_indices(term, dim, dtype_idx, True)
+        m = scipy.sparse.coo_matrix((data, (row, col)), shape=(dim, dim))
+        if sparse is None:
+            return m.toarray()
+        return _sparse_types[sparse](m)
 
 
 _ExprTuple = namedtuple("_ExprTuple", "terms")
@@ -571,16 +650,14 @@ class Expr(_ExprTuple):
         """Make new Expr from a number"""
         if num:
             return Expr.from_term(Term((), num))
-        else:
-            return Expr.zero()
+        return Expr.zero()
 
     @staticmethod
     def from_term(term):
         """Make new Expr from a Term"""
         if term.coeff:
             return Expr((term,))
-        else:
-            return Expr.zero()
+        return Expr.zero()
 
     @staticmethod
     def from_terms_iter(terms):
@@ -745,8 +822,21 @@ class Expr(_ExprTuple):
         return self
 
     def max_n(self):
-        """Returns the maximum index of Pauli matrices in the Term."""
-        return max(term.max_n() for term in self.terms if term.ops)
+        """Returns the maximum index of Pauli matrices in the Expr.
+        If Expr is empty or only constant and identity matrix, returns -1.
+        """
+        try:
+            return max(term.max_n() for term in self.terms if term.ops)
+        except ValueError:
+            return -1
+
+    @property
+    def n_qubits(self):
+        """Returns the number of qubits of the Term.
+
+        If Expr is empty or only constant and identity matrix, returns 0.
+        """
+        return self.max_n() + 1
 
     def coeffs(self):
         """Generator which yields a coefficent for each Term."""
@@ -776,9 +866,39 @@ class Expr(_ExprTuple):
 
     def to_matrix(self, n_qubits=-1, *, sparse=None):
         """Convert to the matrix."""
+        if not (sparse is None or sparse in _sparse_types):
+            raise ValueError(f'Unknown sparse format {sparse}.')
         if n_qubits == -1:
-            n_qubits = self.max_n() + 1
-        return sum(term.to_matrix(n_qubits, sparse=sparse) for term in self.terms)
+            n_qubits = self.n_qubits
+        if n_qubits == 0:
+            m = np.array([[sum(term.coeff for term in self.terms)]])
+            if sparse is None:
+                return m
+            return _sparse_types[sparse](m)
+        expr = self.simplify()
+        grpkey = lambda pau: sum(1 << op.n for op in pau.ops if op.op in 'XY')
+        dim = 2 ** n_qubits
+        is_csr = sparse == 'csr'
+        gr_terms = [list(g) for _, g in groupby(sorted(expr.terms, key=grpkey), key=grpkey)]
+        n_groups = len(gr_terms)
+        n_vals = n_groups * dim
+        dtype_idx = np.int32 if n_qubits < 31 else np.int64
+        vals = np.empty(n_vals, dtype=complex)
+        inds = np.empty(n_vals, dtype=dtype_idx)
+        for i_grp, grp in enumerate(gr_terms):
+            val_acc = _term_to_dataarray(grp[0], n_qubits, is_csr)
+            inds[i_grp::n_groups] = _term_to_indices(grp[0], dim, dtype_idx, False)
+            for term in grp[1:]:
+                val_acc += _term_to_dataarray(term, n_qubits, is_csr)
+            vals[i_grp::n_groups] = val_acc
+        if not is_csr:
+            m = scipy.sparse.csc_matrix((vals, inds, np.arange(0, n_vals + 1, n_groups)), shape=(dim, dim))
+        else:
+            m = scipy.sparse.csr_matrix((vals, inds, np.arange(0, n_vals + 1, n_groups)), shape=(dim, dim))
+        m.eliminate_zeros()
+        if sparse is None:
+            return m.toarray()
+        return _sparse_types[sparse](m)
 
 
 def qubo_bit(n):
